You duty is to decide which tool to use to answer the question.
The question is:

{{ question }}

{{ called_tools }}

{{ tool_prompt }}

You must answer in the following format:


<Thoughts>
Thoughts:
- ...
- ...
- ...
</Thoughts>

<Selected tools>
```json
[
        {
            "name": "tool_name",
            "arguments": {
                "argument_name": "argument_value"
            }
        },
        ...
]
```
</Selected tools>

Note:
- You can leave the selected tools empty if you think none of the tools are relevant to the question or given contexts are enough to answer the question.
- You can select multiple tools if you think multiple tools are relevant to the question.
- You `tool_name` must be one of the tool names in the tools list, and you must spell it correctly.
- If there is no need to call any tool, Please give your answer by continuing the following format:

<Answer>
<Your answer here>
</Answer>

