# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: asr_data_frames.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61sr_data_frames.proto\x12\x0f\x61\x63hatbot_frames\"\xea\x01\n\x19\x41SRLiveTranscriptionFrame\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04text\x18\x03 \x01(\t\x12\x0f\n\x07user_id\x18\x04 \x01(\t\x12\x11\n\ttimestamp\x18\x05 \x01(\t\x12\x10\n\x08language\x18\x06 \x01(\t\x12\x12\n\ntimestamps\x18\x07 \x03(\x04\x12\x11\n\tspeech_id\x18\x08 \x01(\x04\x12\x10\n\x08is_final\x18\t \x01(\x08\x12\x12\n\nstart_at_s\x18\n \x01(\x02\x12\x10\n\x08\x63ur_at_s\x18\x0b \x01(\x02\x12\x10\n\x08\x65nd_at_s\x18\x0c \x01(\x02\"\xab\x01\n\x12TranscriptionFrame\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04text\x18\x03 \x01(\t\x12\x0f\n\x07user_id\x18\x04 \x01(\t\x12\x11\n\ttimestamp\x18\x05 \x01(\t\x12\x10\n\x08language\x18\x06 \x01(\t\x12\x11\n\tspeech_id\x18\x07 \x01(\x04\x12\x12\n\nstart_at_s\x18\x08 \x01(\x02\x12\x10\n\x08\x65nd_at_s\x18\t \x01(\x02\"\x9c\x01\n\x05\x46rame\x12L\n\x16\x61sr_live_transcription\x18\x04 \x01(\x0b\x32*.achatbot_frames.ASRLiveTranscriptionFrameH\x00\x12<\n\rtranscription\x18\x05 \x01(\x0b\x32#.achatbot_frames.TranscriptionFrameH\x00\x42\x07\n\x05\x66rameb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'asr_data_frames_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_ASRLIVETRANSCRIPTIONFRAME']._serialized_start=43
  _globals['_ASRLIVETRANSCRIPTIONFRAME']._serialized_end=277
  _globals['_TRANSCRIPTIONFRAME']._serialized_start=280
  _globals['_TRANSCRIPTIONFRAME']._serialized_end=451
  _globals['_FRAME']._serialized_start=454
  _globals['_FRAME']._serialized_end=610
# @@protoc_insertion_point(module_scope)
