from .qwen2_mtp_v4_48_3 import Qwen2MTPConfig, Qwen2MTPForCausalLM
from .qwen2_mtp_sensevoice_v4_48_3 import Qwen2MTPSenseVoiceConfig, Qwen2MTPSenseVoiceForCausalLM


# from .qwen2_v4_48_3 import Qwen2Config, Qwen2ForCausalLM

