from achatbot.services.a2a_multiagents import ADKBaseHostAgent


class ADKPlanerAgent(ADKBaseHostAgent):
    f"""
    https://github.com/weedge/doraemon-nb/blob/main/langchain/plan_and_execute.ipynb
    !TODO
    """
